"""Model classes to represent layouts and its components."""

import os
import logging

from datalad import api

from typing import List
from typing import Dict

from sqlalchemy import inspect
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import relationship
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import reconstructor
from sqlalchemy.orm import attribute_keyed_dict
from sqlalchemy.orm import DeclarativeBase

from .specification import Specification

__all__ = ["Tag", "File", "Layout"]


class Base(DeclarativeBase):
    """Represents base object of ORM."""

    @classmethod
    def get_existing(cls, *primary, indexer):
        """Returns existing record as ORM object if present."""

        obj = indexer.get(cls, primary)
        return obj


class Tag(Base):
    """Generic tag representation."""

    __tablename__ = "tags"

    path: Mapped[str] = mapped_column(ForeignKey("files.path"), primary_key=True)
    name: Mapped[str] = mapped_column(primary_key=True)
    value: Mapped[str]

    files: Mapped["File"] = relationship(back_populates="tags")

    def __init__(self, path, name, value):
        self.path = path
        self.name = name
        self.value = value

    def __repr__(self):
        return f"<Tag {self.name}:'{self.value}'>"


class File(Base):
    """Generic file representation."""

    __tablename__ = "files"

    path: Mapped[str] = mapped_column(primary_key=True)
    root: Mapped[str] = mapped_column(ForeignKey("layouts.root"))

    layout: Mapped["Layout"] = relationship(back_populates="files")
    tags: Mapped[Dict[str, "Tag"]] = relationship(
        collection_class=attribute_keyed_dict("name"),
        cascade="all, delete-orphan",
    )

    def __init__(self, path, root):
        self.path = path
        self.root = root

    @property
    def rel_path(self):
        r = os.path.relpath(self.path, self.root)
        r = os.path.join(self.layout.prefix, r) if self.layout else r
        return r

    def add_tag(self, name, value):
        """Add a tag to file."""
        self.tags[name] = Tag(self.path, name, value)

    def add_tags(self, **tags):
        """Add tags to file. If tag present, value is updated."""
        for name, value in tags.items():
            self.add_tag(name, value)

    def add_tag_from_func(self, name, func):
        """Add a tag to the file based on func return value."""
        self.tags[name] = Tag(self.path, name, func(self.path))

    def build_path(self):
        """Build path based on tags associated with file."""
        t = self.get_tags()
        return self.layout.spec.build_path(t)

    def build_modified_path(self, **changes):
        "Returns the path for file given changes to tags."
        t = self.get_tags().update(changes)
        return self.layout.spec.build_path(t)

    def get(self):
        """Get file from remote dataset."""
        api.get(self.path, dataset=self.root)

    def get_tags(self):
        """Returns tags associated with the file."""
        return {n: t.value for n, t in self.tags.items()}

    def __repr__(self):
        return f"<File path='{self.rel_path}'>"


class Layout(Base):
    """Representation of file collection in a directory."""

    __tablename__ = "layouts"

    root: Mapped[str] = mapped_column(primary_key=True)
    url: Mapped[str] = mapped_column(nullable=True)

    files: Mapped[List["File"]] = relationship(
        back_populates="layout", cascade="all, delete-orphan"
    )

    def __init__(self, root, url):
        self.root = root
        self.url = url

    @reconstructor
    def _init_on_load(self):
        self.kind = self._decipher_kind()
        self.prefix = self._decipher_prefix()

    @staticmethod
    def _get_indexer(indexer):
        from .indexer import Indexer

        if not indexer:
            indexer = Indexer()
        return indexer

    def _decipher_kind(self):
        for k in ["sourcedata", "derivatives"]:
            if k in self.root:
                return k
        return "primary"

    def _decipher_prefix(self):
        return self.root[p:] if (p := self.root.rfind(self.kind)) != -1 else ""

    def _setup_support(self, indexer, specification):
        self.indexer = indexer
        self.spec = specification

    @staticmethod
    def create(
        root,
        url=None,
        indexer=None,
        spec=Specification(),
        reindex=False,
        valid_only=True,
        from_url=False,
    ):
        """Factory method for Layout.

        Parameters
        ----------
        root : str
            Path of root directory of data.

        url : str
            Datalad supported URL for remote residence.

        indexer : almirah.Indexer
            db to use for indexing.

        spec : almirah.Specification
            Specification to apply.

        reindex : bool, default False
            If True, force reindexing of Layout.

        valid_only : bool, default True
            If False, all contents of indexed irrespective of spec abidance.

        from_url : bool, default False
            If True, clone data from url to populate layout.

        Returns
        -------
        layout: almirah.Layout
            If layout has already been indexed before, a reference to
            the indexed Layout is returned. Else, a new Layout
            instance is indexed and returned.

        """

        indexer = Layout._get_indexer(indexer)
        layout = Layout.get_existing(root, indexer=indexer)

        if not layout:
            layout = Layout(root, url)
            layout._init_on_load()

        layout._setup_support(indexer, spec)
        indexer.add(layout)

        if from_url:
            layout.clone_from_url()

        if reindex or not inspect(layout).persistent:
            layout.reindex(valid_only)

        return layout

    def clone_from_url(self):
        """Clone layout from a datalad url."""
        logging.info(f"Cloning layout from {self.url}")
        api.clone(source=self.url, path=self.root)

    def get_files(self, **filters):
        """Return files that match criteria."""
        files_filter = self.indexer._files_filter(self.root, **filters)
        files = self.indexer.run_query(files_filter)
        return files

    def get_tag_values(self, name, **filters):
        """Return all values available for a tag name."""
        tag_values = self.indexer._tag_values(self.root, name, **filters)
        values = self.indexer.run_query(tag_values)
        return values

    def reindex(self, valid_only=True):
        """Reindex layout to include additions to the layout."""
        logging.info(f"Reindexing {self}")
        self.indexer(self, valid_only)

    def __repr__(self):
        return f"<Layout root='{self.root}'>"



"""Indexing functionality for metadata search and filtering."""

import os
import logging

from sqlalchemy import select
from sqlalchemy import func
from sqlalchemy import tuple_

from .layout import Base, File, Tag
from ..db import DBManager


class Indexer:
    """Index files in a Layout."""

    def __init__(self, db_manager=None):
        self.db = db_manager if db_manager else DBManager()
        Base.metadata.create_all(self.db.engine)

    def __call__(self, layout, valid_only=True):
        logging.info(f"Indexing layout with root {layout.root}")
        self.layout = layout
        self.valid_only = valid_only
        self._index_dir(self.layout.root)

    def _merge(self, obj):
        return self.db.session.merge(obj)

    def _index_dir(self, dir):
        """Iteratively index all directories in layout."""
        logging.info(f"Indexing directory {dir}")
        SKIP_DIRS = ["sourcedata", "derivatives"]
        for content in os.listdir(dir):
            if content in SKIP_DIRS:
                logging.info(f"Skipping content {content}")
                continue

            path = os.path.join(dir, content)
            if not (self._index_file(path) and self.valid_only) and os.path.isdir(path):
                self._index_dir(path)

        self.db.session.commit()

    def _index_file(self, path):
        """Add valid files to index and return true if successful."""
        logging.info(f"Indexing file at {path}")
        file = File(path, self.layout.root)
        if self.valid_only and not self.layout.spec.validate(
            os.path.join(self.layout.prefix, file.rel_path)
        ):
            logging.info("Skipping invalid file")
            return False
        file = self._merge(file)
        self.layout.files.append(file)
        self._index_tags(file)
        self.add(file)
        return True

    def _index_tags(self, file):
        """Add tags of file to index."""
        logging.info("Adding file tags")
        file.add_tag_from_func("is_dir", os.path.isdir)
        tags = self.layout.spec.extract_tags(file.rel_path)
        file.add_tags(**tags)

    def _tags_filter(self, **filters):
        # Unpack filters dict to tuple
        tag_reqs = [(name, value) for name, value in filters.items()]
        logging.debug(f"Tags required: {tag_reqs}")

        # Construct table of passing file paths
        tags_filter = select(Tag.path)

        # If filters present
        if filters:
            tags_filter = (
                tags_filter.where(tuple_(Tag.name, Tag.value).in_(tag_reqs))
                .group_by(Tag.path)
                .having(func.count(Tag.name) == len(tag_reqs))
            )

        return tags_filter

    def _files_filter(self, root, **filters):
        tags_filter = self._tags_filter(**filters).subquery()

        # Build File objects from file paths
        files_filter = (
            select(File)
            .where(File.path.in_(select(tags_filter)))
            .where(File.root == root)
        )

        return files_filter

    def _tag_values(self, root, name, **filters):
        files_filter = self._files_filter(root, **filters).subquery()

        # Build Tag values from file paths
        tag_values = (
            select(Tag.value)
            .where(Tag.path.in_(select(files_filter.c.path)))
            .where(Tag.name == name)
            .distinct()
        )

        return tag_values

    def add(self, obj):
        self.db.session.add(obj)

    def get(self, cls, primary):
        """Get record as an object provided its ORM class and primary key."""
        return self.db.session.get(cls, primary)

    def run_query(self, query):
        """Run a query on db associated and return all results."""
        return self.db.session.scalars(query).all()



"""Module for easy report generation and data summary."""

from sqlalchemy import select, distinct, func

from .layout import File, Tag


class Report:
    """Collection of report generating methods."""

    @staticmethod
    def generate(layout, level="summary", tags=["subject"]):
        """Generate report for a layout."""
        sess = layout.indexer.db.session
        tag_counts = []
        for tag in tags:
            count = sess.scalar(
                select(func.count(distinct(Tag.value)))
                .where(Tag.name == tag)
                .where(File.root == layout.root)
                .join(File)
            )
            tag_counts.append(count)

        if level == "lengthy":
            raise NotImplementedError

        # Generate report string
        print(f"Layout ... name: {layout.name}")
        for index, tag in enumerate(tags):
            print(f"\t | {tag} \t: {tag_counts[index]}")

    @staticmethod
    def generate_for_dataset(dataset):
        """Generate report for a dataset."""

        print(f"Dataset ... {dataset.primary.name}")

        sources = set()
        for s in dataset.sourcedata.layouts:
            sources.add(s.name)
        print(f"Sources | {','.join(sources)}")

        ders = set()
        for d in dataset.derivatives.layouts:
            ders.add(d.name)
        print(f"Pipelines | {','.join(ders)}")

        for lay in dataset.layouts:
            Report.generate(lay)



"""Model classes to represent the dataset."""

import os
from itertools import chain
from abc import ABC, abstractmethod

from .layout import Layout


class Dataset(ABC):
    """Representation of a dataset.

    Attributes
    ----------
    layout_names : list of str
        Name of all the layouts in the dataset.

    Parameters
    ----------
    layouts : list of Layout
        All layouts under the dataset.
    """

    def __init__(self, layouts=list()):
        self.layouts = layouts
        self.layout_names = None
        self._update_layout_names()

    @abstractmethod
    def _build_initialiser_args(root):
        ...

    @classmethod
    def _build_dataset(cls, args):
        return cls(args)

    @classmethod
    def build_from_root(cls, root, indexer=None):
        """Build dataset with the directory provided as root."""
        if not os.path.exists(root):
            print("Path does not exist")
            return
        args = cls._build_initialiser_args(root, indexer)
        return cls._build_dataset(args)

    def add_layout(self, layout):
        """Add a layout to the dataset."""
        if layout.root in [r.root for r in self.layouts]:
            print("Layout already part of dataset")
            return
        self.layouts.append(layout)
        self._update_layout_names()

    def get_files(self, **filters):
        """Return files that match criteria."""
        files = list()
        for lay in self.layouts:
            files.extend(lay.get_files(filters))
        return files

    def _update_layout_names(self):
        self.layout_names = [lay.name for lay in self.layouts] if self.layouts else None

    def __repr__(self):
        classname = self.__class__.__name__
        if not self.layout_names:
            return f"<{classname} layouts='empty'>"
        elif len(self.layout_names) < 4:
            return f"<{classname} layouts={self.layout_names}>"
        else:
            return f"<{classname} layouts=[{self.layout_names[0]}...{self.layout_names[-1]}]>"


class SourceDatasets(Dataset):
    """Representation of the collection of sources."""

    def _build_initialiser_args(root, indexer):
        return [
            Layout(r.path, name=r.name, indexer=indexer)
            for r in os.scandir(root)
            if r.is_dir()
        ]


class DerivativeDatasets(Dataset):
    """Representation of the collection of derivatives."""

    def _build_initialiser_args(root, indexer):
        return [
            Layout(r.path, name=r.name, indexer=indexer)
            for r in os.scandir(root)
            if r.is_dir()
        ]


class CompleteDataset(Dataset):
    """Representation of the complete dataset.

    Attributes
    ----------
    layout_names : list of str
        Name of all the layouts in the dataset.

    Parameters
    ----------
    primary : Layout
        Layout of clean and curated data in dataset.
    sourcedata : SourceDatasets or list of Layout
        Sources for the primary.
    derivatives : DerivativeDatasets or list of Layout
        Data derived from primary.
    """

    def __init__(self, primary, sourcedata=None, derivatives=None):
        if isinstance(sourcedata, list):
            sourcedata = SourceDatasets(sourcedata)

        if isinstance(derivatives, list):
            derivatives = DerivativeDatasets(derivatives)

        self.primary = primary
        self.sourcedata = sourcedata
        self.derivatives = derivatives
        self._update_layouts()

    def _build_initialiser_args(root, indexer):
        return (
            Layout(root, name=os.path.basename(root), indexer=indexer),
            SourceDatasets.build_from_root(
                os.path.join(root, "sourcedata"), indexer=indexer
            ),
            DerivativeDatasets.build_from_root(
                os.path.join(root, "derivatives"), indexer=indexer
            ),
        )

    def _build_dataset(args):
        return CompleteDataset(*args)

    def _update_layouts(self):
        """Update list of layouts under dataset."""
        primary_list = [self.primary]
        self.layouts = list(
            chain(
                primary_list,
                self.sourcedata.layouts,
                self.derivatives.layouts,
            )
        )
        self.layout_names = None
        self._update_layout_names()

    def add_derivative(self, layout):
        """Add a derivative to the dataset."""
        self.derivatives.add_layout(layout)
        self._update_layouts()

    def add_source(self, layout):
        """Add a source to the dataset."""
        self.sourcedata.add_layout(layout)
        self._update_layouts()



"""Utility functions."""

import os
import re
import yaml
import shutil
import logging
import subprocess

from functools import reduce
from os.path import dirname as up


def copy(src, dst, overwrite=False):
    """Copies from source to destination."""
    if not dst:
        raise TypeError("Expected destination path, received None")

    if os.path.exists(dst):
        if not overwrite:
            logging.warning("Skipping copy as file exists")
            return
        if overwrite:
            logging.warning("Overwriting existing file")
            remover = shutil.rmtree if os.path.isdir(dst) else os.remove
            remover(dst)

    os.makedirs(os.path.dirname(dst), exist_ok=True)
    copier = shutil.copytree if os.path.isdir(src) else shutil.copy2
    copier(src, dst)


def get_dir_contents(root, pattern, skip=None):
    """Return list of contents in a directory that match pattern."""
    m = []
    for dir, subdirs, files in os.walk(root):
        contents = subdirs + files
        for content in contents:
            if re.match(pattern, content):
                m.append(os.path.join(dir, content))
    matches = [c for c in m if not any([re.search(s, c) for s in skip or []])]
    return matches


def deep_get(dictionary, keys, default=None):
    return reduce(
        lambda d, key: d.get(key, default) if isinstance(d, dict) else default,
        keys.split("."),
        dictionary,
    )


def read_yaml(path):
    """Return dict equivalent of yaml file."""
    with open(path) as f:
        content = yaml.load(f, yaml.SafeLoader)
        return content


def read_yamls(path):
    """Returns list of dict equivalents of yamls in a file."""
    with open(path) as f:
        content = yaml.load_all(f, yaml.SafeLoader)
        return list(content)


def run_shell(cmd, suppress_output=True):
    """Execute shell command in background."""
    sp = subprocess.run(cmd, shell=True)
    return sp


def log_df(df, msg, level=logging.ERROR, **kwargs):
    """
    Log df records with a message.

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame to be logged.

    msg: str
        Log message compatible with string formatting.

    level: int, optional
        Logging level to use. Accepts logging.LEVEL values.

    kwargs: key, value mappings
        Other keyword arguments are passed to `str.format()`.
    """
    if not df.empty:
        logging.log(level, msg.format(df=df.to_string(), **kwargs))


def get_dtype(dtype, default_length=250):
    """Returns supported dtype and length from provided dtype string."""

    SUPPORTED_DTYPES = {"boolean", "datetime", "float", "integer", "str"}

    dtype, length = re.match(r"(\w+).?(\d+)?", dtype).groups()

    if dtype not in SUPPORTED_DTYPES:
        raise TypeError(f"Unsupported dtype {dtype} encountered")

    if length and dtype != "str":
        raise ValueError(f"Expected zero, but received arg for dtype {dtype}")

    if not length and dtype == "str":
        length = default_length

    length = int(length) if length else None
    return dtype, length


def get_tutorial_dataset(dst):
    """Copies the tutorial dataset to destination."""

    import almirah

    path = os.path.join(up(up(almirah.__file__)), "tests/data/tutorial")
    copy(path, dst)



from .report import Report
from .convert import convert
from .indexer import Indexer
from .layout import File, Layout, Tag
from .specification import Specification

__all__ = [
    "convert",
    "File",
    "Layout",
    "Indexer",
    "Tag",
    "Specification",
    "Report",
]
